/***    cctool.c    ****/

/**********************************************************************
*  Copyright (c) 1990, 1991, 1992 Iris Computing Laboratories.
*
*  This software is provided for demonstration purposes only.  As
*  freely-distributed, modifiable source code, this software carries
*  absolutely no warranty.  Iris Computing Laboratories disclaims
*  all warranties for this software, including any implied warranties
*  of merchantability and fitness, and shall not be liable for
*  damages of any type resulting from its use.
*  Permission to use, copy, modify, and distribute this source code
*  for any purpose and without fee is hereby granted, provided that
*  the above copyright and this permission notice appear in all copies
*  and supporting documentation, and provided that Iris Computing
*  Laboratories not be named in advertising or publicity pertaining
*  to the redistribution of this software without specific, written
*  prior permission.
**********************************************************************/

/***************************************************
cctool.c provides lower level routines for xcctool.
***************************************************/

#include <ctype.h>
#include <memory.h>
#include <string.h>

#include "cctool.h"
#include "alert.h"
#include "editor.h"

/***********************************************************************
We chose to create the temporary file in the current directory, instead
of in `/tmp' (with a pid attached) because it is not automatically
deleted after its initial use.  That is, the user may activate the
error window several times to re-review errors between compilations.
***********************************************************************/

/*
Private functions:
*/

static void process_make_errors(); 
static int save_before_compile();
static int switch_edit_files();
static char *compose_lib_options();
static char *get_filename();
static char *get_make_filename();
static char *get_compile_filename();
static char *get_current_working_directory();
static int create_file();
static int file_exists_alert();


/*
Private globals:
*/

static int line_nums[MAX_ERR];			/* used for error messages	*/
static char err_msgs[MAX_ERR][MAX_STR];	/* during compilation		*/
static char err_files[MAX_ERR][MAX_STR];
static int next_err, previous_err;

static char current_edit_file[MAX_PATH_LEN + MAX_FILENAME_LEN];


/*
initialize_cctool_structures() sets up all data required by
routines in cctool.c
*/

void initialize_cctool_structures()
{
	*current_edit_file = EOS;
}	/* initialize_cctool_structures */


/*
compile_file() builds a command-line string to invoke the C compiler.
compose_lib_options() is used to supply libraries for linking.  Output
from the C compiler is redirected to the file ".xcctool.cc.stderr."  If
there are errors, the compilation-error window is displayed.
*/

void compile_file()
{
	char *filename;
	char cmd_string[MAX_PATH_LEN + MAX_FILENAME_LEN + MAX_STR];
	char error_spec[MAX_PATH_LEN + MAX_FILENAME_LEN + 17];

	if ((filename = get_filename(filenameBoxCompile)) == NULL)
		return;
	if (!save_before_compile())
		return;
	sprintf(error_spec, "%s/%s",
		get_current_working_directory(), ".xcctool.cc.stderr");
	unlink(error_spec);
	filename = get_compile_filename();
	sprintf(cmd_string, "%s %s %s %s 2> %s",
		XiSimpleTextGetString(compilerBoxCompile),
		XiSimpleTextGetString(optionsBoxCompile),
		filename,
		compose_lib_options(),
		error_spec);
	system(cmd_string);
	next_err = 0; 			/* start fresh for each compilation */
	previous_err = -1;
	retrieve_errors(error_spec);
	if (line_nums[0]) {		/* anything other than 0 implies errors */
		error();
		process_next_error();
	}
	else
		XiSimpleTextSetString(errorMsgBox,
			"*** no compilaton errors ***");
}	/* compile_file */


/*
retrieve_errors() builds an array of line numbers for each error and
an array of corresponding messages.  Both arrays, plus the variable
next_err, are global.  The information for these arrays is built from
the file ".xcctool.cc.stderr", which contains the redirected output from
the C compiler.
*/

void retrieve_errors(error_spec)
char *error_spec;
{
	FILE *error_fd;
	char *eof_check, line[MAX_STR * 2], line_number[12], terminator;
	int i, j, num_errs = 0, no_file_spec;

 	line_nums[0] = 0;					/* 0 indicates end of messages */
 	if ((error_fd = fopen(error_spec, "r")) == NULL) {
 		alert("Can't open temporary error-reporting file.");
		return;
	}
	while (((eof_check = fgets(line, MAX_STR * 2, error_fd)) != NULL) &&
			num_errs < MAX_ERR - 2) {
		no_file_spec = FALSE;
		terminator = (line[0] == '"') ? '"' : ':';
		if (terminator != '"') {	/* scan for '.' in file spec. */
			for (i = 0;
					line[i] && line[i] != terminator && line[i] != '.'; i++)
				;
			if (line[i] != '.') {				/* no file spec.,		*/
				err_files[num_errs][0] = EOS;	/* so get entire line	*/
				i = 0;
				line_nums[num_errs] = 1;
				no_file_spec = TRUE;
			}
		}
		if (!no_file_spec) {
			i = (line[0] == '"') ? 1 : 0;
			for (j = 0; line[i] && line[i] != terminator; i++, j++)
				err_files[num_errs][j] = line[i];	/* extract file spec. */
			err_files[num_errs][j] = EOS;
			for (j = i ; line[i] && !isdigit(line[i]); i++)	/* skip junk */
				;
			if (!line[i]) {							/* no line number */
				i = j;
				while (line[i] == terminator || line[i] == ' ')
					i++;
				line_nums[num_errs] = NO_LINE_NUM;
			}
			else {
				for (j = 0; line[i] && isdigit(line[i]) &&
						j < 12; i++, j++)				/* extract line no. */	
					line_number[j] = line[i];
				line_number[j] = EOS;
				line_nums[num_errs] = atoi(line_number);
				i += 2;									/* skip junk */
			}
		}
		for (j = 0; line[i] && j < MAX_STR; i++, j++)
			err_msgs[num_errs][j] = line[i];		/* extract message */
		if (err_msgs[num_errs][j - 1] == LF)
			err_msgs[num_errs][j - 1] = EOS;
		else
			err_msgs[num_errs][j] = EOS;
		num_errs++;
	}
	if (eof_check != NULL) {
		strcpy(err_msgs[num_errs],
				"*** maximum number of errors exceeded ***");
		line_nums[num_errs++] = -1;
	}
	line_nums[num_errs] = 0;			/* 0 indicates end of messages */
	fclose(error_fd);
}	/* retrieve_errors */


/*
process_next_error() displays the next error from the error
array, and then increments the index to the error messages.
*/

void process_next_error()
{
	char file_spec[MAX_PATH_LEN + MAX_FILENAME_LEN + 30];

	if (!line_nums[next_err]) {
		XiSimpleTextSetString(errorMsgBox, "*** end of errors ***");
		previous_err = next_err - 1;
		return;
	}
	if (strlen(err_files[next_err]) == 0)	/* no specific file */
		XiSimpleTextSetString(errorMsgBox, err_msgs[next_err]);
	else {				/* a specific file has been referenced */
		if (!file_exists_alert(err_files[next_err])) {
 			sprintf(file_spec,
 				"!!!source file not found: %s", err_files[next_err]);
			XiSimpleTextSetString(errorMsgBox, file_spec);
			return;
		}
		if (strcmp(current_edit_file, err_files[next_err]))
			if (!switch_edit_files(err_files[next_err]))
				return;
		XiSimpleTextSetString(errorMsgBox, err_msgs[next_err]);
		if (line_nums[next_err] != NO_LINE_NUM)
			goto_line_edit_buffer(line_nums[next_err] - 1);
	}
	previous_err = next_err - 1;
	next_err++;
}	/* process_next_error */


/*
process_previous_error() displays the previous error from the
error array by setting next_err to the value of previous_err,
before calling process_next_error().
*/

void process_previous_error()
{
	if (next_err == 0 || previous_err < 0)
		return;
	next_err = previous_err;
	process_next_error();
}	/* process_previous_error */


/*
make_file() builds a command-line string to invoke the UNIX make facility.
If the user selects "Yes" in the make window for "Pass Libraries to
$(LDFLAGS)?", the libraries specified in the main window are passed to the
linker.
*/

void make_file()
{
	char cmd_string[MAX_PATH_LEN + MAX_FILENAME_LEN + MAX_STR],
		file_spec[MAX_PATH_LEN + MAX_FILENAME_LEN],
		stderr_spec[MAX_PATH_LEN + MAX_FILENAME_LEN + 19],
		stdout_spec[MAX_PATH_LEN + MAX_FILENAME_LEN + 19];
	char *ldflags_str = XiSimpleTextGetString(ldflagsBoxMake);
	int cflags = (XiChoiceGetSelectedItem(cflagsBoxMake) == YES);
	int libraries = (XiChoiceGetSelectedItem(librariesBoxMake) == YES);
	int ldflags = (strlen(ldflags_str) > 0);

	if (ldflags) {	/* is it an empty string */
		int i;
		for (i = 0; ldflags_str[i] && index(" \t", ldflags_str[i]); i++)
			;
		if (!ldflags_str[i]) /* whitespace only */
			ldflags = FALSE;
	}
	sprintf(stderr_spec, "%s/%s",
		get_current_working_directory(), ".xcctool.make.stderr");
	sprintf(stdout_spec, "%s/%s",
		get_current_working_directory(), ".xcctool.make.stdout");
	unlink(stderr_spec);			/* unlink them so that the I/O	*/
	unlink(stdout_spec);			/* redirections will work OK	*/
	if (get_make_filename() == NULL)
		*file_spec = EOS;
	else
		sprintf(file_spec, "-f %s", get_make_filename()); 
	sprintf(cmd_string,
		"make %s %s %s %s%s %s %s %s%s > %s 2> %s",
		file_spec,
		XiSimpleTextGetString(optionsBoxMake),
		cflags ? "\"CFLAGS= " : "",
		cflags ? XiSimpleTextGetString(optionsBoxCompile) : "",
		cflags ? "\"" : "",
		(libraries || ldflags) ? "\"LDFLAGS= " : "",
		ldflags ? ldflags_str : "",
		libraries ? compose_lib_options() : "",
		(libraries || ldflags) ? "\"" : "",
		stdout_spec,
		stderr_spec);
/*	printf("make command string:::%s:::\n", cmd_string);*/
	system(cmd_string);
	process_make_errors(stderr_spec);
}	/* make_file */

	
/*
process_make_errors() checks for make errors, displaying
the error output from make(1) in an alert box.
*/

static void process_make_errors(stderr_spec)
char *stderr_spec;
{
	char *make_buffer, *output_preview;
	int makefile_size;

	if ((makefile_size = (int) file_size(stderr_spec)) > 0) {
		if ((make_buffer = (char *) malloc((unsigned)(makefile_size + 1))) == NULL) {
			alert("Memory allocation error during make operations!!");
			return;
		}
		if (load_buffer_with_text(stderr_spec, make_buffer)) {
			alert("\"make\" output in .xcctool.make.stderr");
			output_preview = strtok(make_buffer, "\n\r\t");
			if (output_preview != NULL)
				alert(output_preview);	
		}
		free(make_buffer);
	}
}	/* process_make_errors */


/*
save_before_compile() checks for unsaved changes to the edit file
before beginning the compilation.
*/

static int save_before_compile()
{
	char msg[MAX_PATH_LEN + MAX_FILENAME_LEN + 30];
	int result;

	if (!modified_edit_buffer())
		return TRUE;
	edit();
	sprintf(msg, "Save changes to edit file (%s) before compiling?",
		current_edit_file);
	result = alert_query(msg);
	if (result == ALERT_CANCEL) {
		return FALSE;
	}
	if (result == ALERT_YES)
		save_edit_buffer(current_edit_file);
	return TRUE;
}	/* save_before_compile */


/*
compose_lib_options() builds "cc" options for library linkage by
concatenating "-l" to each library specified in the cc options window.
*/

static char *compose_lib_options()
{
	static char lib_options[MAX_STR], libs[MAX_STR];
	char *next;

	strcpy(lib_options, "");
	strncpy(libs, XiSimpleTextGetString(librariesBoxCompile), MAX_STR - 1);
	next = strtok(libs, " \t");
	while (next != NULL &&
			((strlen(lib_options) + strlen(next) + 3) < (MAX_STR - 1))) {
		strcat(lib_options, "-l");
		strcat(lib_options, next);
		strcat(lib_options, " ");
		next = strtok(NULL, " \t");
	}
	return lib_options;
}	/* compose_lib_options */


/*
get_make_filename() gets a filename based on the value
given in the make options window.  If no makefile is specified
it returns NULL.
*/

static char *get_make_filename()
{
	static char file_spec[MAX_PATH_LEN + MAX_FILENAME_LEN];

	sprintf(file_spec, "%s%s%s",
		XiSimpleTextGetString(directoryBoxMake),
		(!strlen(XiSimpleTextGetString(directoryBoxMake))) ?
			"" : "/",
		XiSimpleTextGetString(filenameBoxMake));
	if (!strlen(file_spec))
		return NULL;
	else
		return file_spec;
}	/* get_make_filename */


/*
get_compile_filename() gets a filename, with path, for the
file in the compile options window.
*/

static char *get_compile_filename()
{
	static char file_spec[MAX_PATH_LEN + MAX_FILENAME_LEN];

	sprintf(file_spec, "%s%s%s",
		XiSimpleTextGetString(directoryBoxCompile),
		(!strlen(XiSimpleTextGetString(directoryBoxCompile))) ?
			"" : "/",
		XiSimpleTextGetString(filenameBoxCompile));
	if (!strlen(file_spec))
		return NULL;
	else
		return file_spec;
}	/* get_compile_filename */


/*
get_filename() gets a filename based on the current
filing mode--there is no path, just a filename.
*/

static char *get_filename(w)
Widget w;
{
	static char *filename;

	if (!strlen(filename = XiSimpleTextGetString(w))) {
		alert("Please enter a value for \"Filename:\".");
		return NULL;
	}
	else
		return filename;
}	/* get_filename */


/*
get_current_working_directory() retrieves and then returns a
pointer to the current working directory for the process.
*/

static char *get_current_working_directory()
{
	static char cwd[MAX_PATH_LEN];

	if (!getwd(cwd))
		return NULL;
	else
		return cwd;
}	/* get_current_working_directory */


/*
file_exists_alert() checks for the existence of a file using stat(),
and prompts the user in an alert box.
*/

static int file_exists_alert(filename)
char *filename;
{
	struct stat	statbuf;
	char msg[MAX_PATH_LEN + MAX_FILENAME_LEN + MAX_STR];

	if (stat(filename, &statbuf) < 0) {
		sprintf(msg, "%s %s", filename, "not found.");
		alert(msg);
		return FALSE;
	}
	return TRUE;
}	/* file_exists_alert */


/*
edit_text() checks file-related details and loads the
edit buffer.
*/

int edit_text()
{
	char *filename;
	int status;

	if ((filename = get_filename(filenameBoxCompile)) == NULL)
		return FALSE;
	filename = get_compile_filename();
	if (!check_save_edit_buffer())
		return FALSE;
	if (!file_exists_alert(filename)) 
		if (!create_file(filename))
			return FALSE; 			/* return if file not found */
									/* and file can't be created */
	if ((status = load_edit_buffer(filename)))
		strcpy(current_edit_file, filename);
	edit();
	return status;
}	/* edit_text */


/*
switch_edit_files() checks for modifications to the current
edit buffer, before switching to a new file.
*/

static int switch_edit_files(new_file_spec)
char *new_file_spec;
{
	if (!check_save_edit_buffer())
		return FALSE;
	if (load_edit_buffer(new_file_spec))
		strcpy(current_edit_file, new_file_spec);
	edit();
	return TRUE;
}	/* switch_edit_files */


/*
check_save_edit_buffer() gives the user a change to save changes to
the edit subwindow.
*/

int check_save_edit_buffer()
{
	char msg[MAX_PATH_LEN + MAX_FILENAME_LEN + 30];
	int result;

	if (!modified_edit_buffer())
		return TRUE;
	sprintf(msg, "Save existing changes to edit file: %s?",
		current_edit_file);
	result = alert_query(msg);
	if (result == ALERT_CANCEL) {
		return FALSE;
	}
	if (result == ALERT_YES)
		save_edit_buffer(current_edit_file);
	return TRUE;
}	/* check_save_edit_buffer */


/*
create_file() is used to create a new file for editing operations.
*/

static int create_file(filename)
char *filename;
{
	FILE *fd;
	char msg[MAX_PATH_LEN + MAX_FILENAME_LEN + 30];
	int result;

	sprintf(msg, "OK to create it: %s?", filename);
	if ((result = alert_query(msg)) == ALERT_YES) {
		if ((fd = fopen(filename, "w")) == NULL) {
			sprintf(msg, "Couldn't create file: %s.", filename);
			alert(msg);
			return FALSE;
		}
		else {
			sprintf(msg, "%s has been created.", filename);
			alert(msg);
			fclose(fd);
			return TRUE;
		}
	}
	else if (result == ALERT_CANCEL)
		return FALSE;
	else
		return FALSE;
}	/* create_file */


/*
load_buffer_with_text() loads the buffer from a disk file.
*/

static int load_buffer_with_text(file_spec, buffer)
char *file_spec, *buffer;
{
	FILE *file_ptr;

	if ((file_ptr = fopen(file_spec, "r")) == NULL)
		return FALSE;
	while ((*buffer = fgetc(file_ptr)) != (char) EOF)
		buffer++;
	*buffer = EOS;
	fclose(file_ptr);
	return TRUE;
}	/* load_buffer_with_text */


/*
save_proc() is the procedure that's registered with the
editor module using set_edit_save_proc().
*/

void save_proc()
{
	save_edit_buffer(current_edit_file);
}	/* save_proc */


