/****    editor.c    ****/

/**********************************************************************
*  Copyright (c) 1990, 1991, 1992 Iris Computing Laboratories.
*
*  This software is provided for demonstration purposes only.  As
*  freely-distributed, modifiable source code, this software carries
*  absolutely no warranty.  Iris Computing Laboratories disclaims
*  all warranties for this software, including any implied warranties
*  of merchantability and fitness, and shall not be liable for
*  damages of any type resulting from its use.
*  Permission to use, copy, modify, and distribute this source code
*  for any purpose and without fee is hereby granted, provided that
*  the above copyright and this permission notice appear in all copies
*  and supporting documentation, and provided that Iris Computing
*  Laboratories not be named in advertising or publicity pertaining
*  to the redistribution of this software without specific, written
*  prior permission.
**********************************************************************/


#include "editor.h"
#include "alert.h"


/*
Globals:
*/


Widget editShell, vPaneEdit, buttonBoxEdit, labelEdit, textEdit,
	buttonDoneEdit, buttonSaveEdit;


/*
Private globals:
*/

static char *edit_buffer;
static int buffer_modified = FALSE;
static int edit_buffer_size = EDIT_MAX_TEXT_LEN;


/*
Private functions:
*/

static int file_exists_alert();
static int buf_line_length(), go_line_start();
static void reset_buffer_modified();
static void BufferModified(), EditSave();


/*
The user of this module can set the editor save function;
see set_edit_save_proc().
*/

static void (*edit_save)() = NULL;


/*
set_edit_save_proc() allows another module to supply an editor`
save function.
*/

void set_edit_save_proc(proc)
void (*proc)();
{
	edit_save = proc;
}	/* set_edit_save_proc */


/*
create_edit_window() creates/initializes the simple widget
that's used to house the edit widget.
*/

void create_edit_window(edit_width, edit_height, edit_size, edit_font)
int edit_width, edit_height, edit_size;
XFontStruct *edit_font;
{
	Arg args[10];
	int i;

	edit_buffer_size = edit_size;
	i = 0;
	XtSetArg(args[i], XtNallowShellResize, (XtArgVal) FALSE); i++;
	XtSetArg(args[i], XtNmappedWhenManaged, (XtArgVal) FALSE); i++;
	editShell = XtCreateApplicationShell("editShell",
		topLevelShellWidgetClass, args, i);
	i = 0;
	XtSetArg(args[i], XtNallowResize, (XtArgVal) FALSE); i++;
	vPaneEdit = XtCreateManagedWidget("vPaneEdit",
#ifdef X11R3
		vPanedWidgetClass, editShell, args, i);
#else
		panedWidgetClass, editShell, args, i);
#endif
	buttonBoxEdit = XtCreateManagedWidget("buttonBoxEdit",
		boxWidgetClass, vPaneEdit, NULL, 0);
	i = 0;
	XtSetArg(args[i], XtNlabel, (XtArgVal) ""); i++;
	XtSetArg(args[i], XtNborderWidth, (XtArgVal) 0); i++;
	XtSetArg(args[i], XtNresizable, (XtArgVal) TRUE); i++;
	labelEdit = XtCreateManagedWidget("labelEdit",
		labelWidgetClass, vPaneEdit, args, i);
	i = 0;
#ifdef X11R3
	edit_buffer = XtMalloc(edit_buffer_size + 1);
	*edit_buffer = EOS;
	XtSetArg(args[i], XtNeditType, (XtArgVal) XttextEdit); i++;
	XtSetArg(args[i], XtNtextOptions, (XtArgVal) scrollVertical); i++;
#else
	edit_buffer = "";
	XtSetArg(args[i], XtNeditType, (XtArgVal) XawtextEdit); i++;
	XtSetArg(args[i], XtNscrollVertical,
		(XtArgVal) XawtextScrollAlways); i++;
#endif
	XtSetArg(args[i], XtNlength, (XtArgVal) edit_buffer_size); i++;
	XtSetArg(args[i], XtNwidth, (XtArgVal) edit_width); i++;
	XtSetArg(args[i], XtNheight, (XtArgVal) edit_height); i++;
	XtSetArg(args[i], XtNfont, (XtArgVal) edit_font); i++;
	textEdit = XtCreateManagedWidget("textEdit",
#ifdef X11R3
		asciiStringWidgetClass,
#else
		asciiTextWidgetClass,
#endif
		vPaneEdit, args, i);
	i = 0;
	XtSetArg(args[i], XtNlabel, (XtArgVal) "Done"); i++;
	buttonDoneEdit = XtCreateManagedWidget("buttonDoneEdit",
		commandWidgetClass, buttonBoxEdit, args, i);
	XtAddCallback(buttonDoneEdit, XtNcallback, EditDone, NULL);
	i = 0;
	XtSetArg(args[i], XtNlabel, (XtArgVal) "Save"); i++;
	buttonSaveEdit = XtCreateManagedWidget("buttonSaveEdit",
		commandWidgetClass, buttonBoxEdit, args, i);
	XtAddCallback(buttonSaveEdit, XtNcallback, EditSave, NULL);
	reset_buffer_modified();
}	/* create_edit_window */


/*
edit() is the starting point for error pop-up operations.
*/

void edit()
{
	XtMapWidget(editShell);
}	/* edit */


/*
edit_done() `un-does' the work of edit().
*/

void edit_done()
{
	XtUnmapWidget(editShell);
}	/* edit_done */


/*
cleanup_edit_window() releases and edit window-related data.
*/

void cleanup_edit_window()
{
#ifdef X11R3
	XtFree(edit_buffer);
#else
#endif
}	/* cleanup_edit_window */


/*
load_edit_buffer() loads the edit buffer with the specified file.
*/

int load_edit_buffer(file_spec)
char *file_spec;
{
	FILE *file_ptr;
	static char *buffer, *buf_ptr;
	char msg[MAX_PATH_LEN + MAX_FILENAME_LEN + 50];
	Arg args[5];
	int i;

	if (!file_exists_alert(file_spec))
		return FALSE;
	if (file_size(file_spec) >= edit_buffer_size) {
		sprintf(msg, "File is too large for edit buffer: %s", file_spec);
		alert(msg);
		return FALSE;
	}
	if ((file_ptr = fopen(file_spec, "r")) == NULL) {
		sprintf(msg, "Can't open file: %s", file_spec);
		alert(msg);
		return FALSE;
	}
#ifdef X11R3
	buffer = edit_buffer;
#else
	buffer = XtMalloc((unsigned) (file_size(file_spec) + 1));
#endif
	buf_ptr = buffer;
	while ((*buf_ptr = fgetc(file_ptr)) != (char) EOF)
		buf_ptr++;
	*buf_ptr = EOS;
#ifdef X11R3
	XtTextSetInsertionPoint(textEdit, (XtTextPosition) 0);
	XtTextSetLastPos(textEdit, (XtTextPosition) strlen(edit_buffer));
	XtTextDisplay(textEdit);
#else
	XtSetArg(args[0], XtNstring, (XtArgVal) buffer);
	XtSetValues(textEdit, args, 1);
	XtFree(buffer);
#endif
	fclose(file_ptr);
	reset_buffer_modified();
	i = 0;
	sprintf(msg, "File: %s", file_spec);
	XtSetArg(args[i], XtNlabel, (XtArgVal) msg); i++;
	XtSetValues(labelEdit, args, i);
	return TRUE;
}	/* load_edit_buffer */


/*
save_edit_buffer() saves the edit buffer to the specified file.
*/

int save_edit_buffer(file_spec)
char *file_spec;
{
	FILE *file_ptr;
	register int i, write_ok = TRUE;
	Arg args[1];
	char *buffer;

#ifdef X11R3
	buffer = edit_buffer;
#else
	XtSetArg(args[0], XtNstring, (XtArgVal) &buffer);
	XtGetValues(textEdit, args, 1);
#endif

	if (*file_spec == EOS) {
		alert("Zero-length filename.");
		return FALSE;
	}
	if ((file_ptr = fopen(file_spec, "w")) == NULL) {
		alert("Can't write the file.");
		return FALSE;
	}
	for (i = 0; buffer[i] && write_ok == TRUE; i++)
		write_ok = (fputc(buffer[i], file_ptr) != (char) EOF);
	if (write_ok != TRUE) {
		alert("Error while writing the file.");
		fclose(file_ptr);
		return FALSE;
	}
	fclose(file_ptr);
	reset_buffer_modified();
	return TRUE;
}	/* save_edit_buffer */


/*
goto_line_edit_buffer() advances to (the beginning of) the specified
line in the edit buffer.
*/

int goto_line_edit_buffer(line)
int line;
{
	Arg args[1];
	char *buffer;
	register int i, lines_moved;

#ifdef X11R3
	buffer = edit_buffer;
#else
	XtSetArg(args[0], XtNstring, (XtArgVal) &buffer);
	XtGetValues(textEdit, args, 1);
#endif
	for (i = lines_moved = 0; buffer[i] && line > 0; ) {
		for ( ; buffer[i] && buffer[i] != LF; i++)
			;
		if (buffer[i]) {
			i++;
			line--;
			lines_moved++;
		}
	}
	i = go_line_start(buffer, i);
#ifdef X11R3
	XtTextSetInsertionPoint(textEdit, i);
	XtTextDisplay(textEdit);
#else
	XawTextSetInsertionPoint(textEdit, i);
	XawTextDisplay(textEdit);
#endif
	return lines_moved;
}	/* goto_line_edit_buffer */


/*
file_size() checks for the existence of a file using stat(),
returning -1 for a nonexistent file and the file size otherwise.
*/

int file_size(file_spec)
char *file_spec;
{
	struct stat	statbuf;

	if (stat(file_spec, &statbuf) < 0)
		return -1;
	else
		return (int) statbuf.st_size;
}	/* file_size */


/*
modified_edit_buffer() is a predicate that indicates whether or not
the edit buffer is currently `unsaved'.
*/

int modified_edit_buffer()
{
	return buffer_modified;
}	/* modified_edit_buffer */


/*
go_line_start() moves the cursor to the beginning of the
line containing the offset `pos' in the buffer.
*/

static int go_line_start(buffer, pos)
char *buffer;
int pos;
{
	if (pos == 0)
		return FALSE;
	else if (buf_line_length(buffer, pos) == 1)
		return pos;
	else {
		if (buffer[pos] == LF)
			pos--;
		for ( ; pos >= 0 && buffer[pos] != LF; pos--)
			;
		pos++;
		return pos;
	}
}   /* go_line_start */


/*
buf_line_length() returns the length of the
current line in the buffer.
*/

static int buf_line_length(buffer, pos)
char *buffer;
int pos;
{
	register int end, start;

	for (end = pos; buffer[end] && buffer[end] != LF; end++)
		;
	for (start = end - 1; start > 0 && buffer[start] != LF; start--)
		;
	if (start == 0)
		return end - start + 1;
	else
		return end - start;
}   /* buf_line_length */


/*
file_exists_alert() checks for the existence of a file,
and prompts the user in an alert box.
*/

static int file_exists_alert(file_spec)
char *file_spec;
{
	char msg[MAX_PATH_LEN + MAX_FILENAME_LEN + 115];

	if (file_size(file_spec) == -1) {
		sprintf(msg, "%s %s", file_spec, "not found.");
		alert(msg);
		return FALSE;
	}
	return TRUE;
}	/* file_exists_alert */


/*
reset_buffer_modified() resets a private variable that records
whether or not the edit buffer is currently `unsaved'.
*/

static void reset_buffer_modified()
{
#ifdef X11R3
	/* doesn't work */XtAddCallback(XtTextGetSource((TextWidget)textEdit), XtNcallback,
		BufferModified, NULL);
#else
	XtAddCallback(XawTextGetSource(textEdit), XtNcallback,
		BufferModified, NULL);
#endif
	buffer_modified = FALSE;
}	/* reset_buffer_modified */


/*
BufferModified() is a callback function that is called to
record the `unsaved' status of the edit buffer.
*/
/*ARGSUSED*/
static void BufferModified(w, client_data, call_data)
Widget w;
caddr_t client_data, call_data;
{
	XtRemoveCallback(w, XtNcallback, BufferModified, NULL);
	buffer_modified = TRUE;
}	/* BufferModified */


/*
EditSave() executes the user-supplied save procedure.
*/
/*ARGSUSED*/
static void EditSave(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
	if (edit_save != NULL)
		edit_save();
}	/* EditSave */


/*
EditDone() removes the `edit' window.
*/
/*ARGSUSED*/
void EditDone(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
	if (modified_edit_buffer()) {
		int result;

		result = alert_query("Save changes to edit file?");
		if (result == ALERT_CANCEL)
			return;
		if (result == ALERT_YES)
			if (edit_save != NULL)
				edit_save();
	}
	edit_done();
}	/* EditDone */

